@@ -1,4 +1,4 @@
-/* tftp-hpa: $Id: tftpd.c,v 1.7 2001/03/30 01:30:50 hpa Exp $ */
+/* tftp-hpa: $Id: tftpd.c,v 1.11 2001/04/23 22:50:52 hpa Exp $ */
 
 /* $OpenBSD: tftpd.c,v 1.13 1999/06/23 17:01:36 deraadt Exp $	*/
 
@@ -41,7 +41,7 @@ static const char *copyright =
  All rights reserved.\n";
 /*static char sccsid[] = "from: @(#)tftpd.c	5.13 (Berkeley) 2/26/91";*/
 /*static char rcsid[] = "$OpenBSD: tftpd.c,v 1.13 1999/06/23 17:01:36 deraadt Exp $: tftpd.c,v 1.6 1997/02/16 23:49:21 deraadt Exp $";*/
-static const char *rcsid = "tftp-hpa $Id: tftpd.c,v 1.7 2001/03/30 01:30:50 hpa Exp $";
+static const char *rcsid = "tftp-hpa $Id: tftpd.c,v 1.11 2001/04/23 22:50:52 hpa Exp $";
 #endif /* not lint */
 
 /*
@@ -53,6 +53,7 @@ static const char *rcsid = "tftp-hpa $Id: tftpd.c,v 1.7 2001/03/30 01:30:50 hpa
 #include <sys/types.h>
 #include <sys/ioctl.h>
 #include <sys/stat.h>
+#include <sys/time.h>
 #include <signal.h>
 #include <fcntl.h>
 
@@ -66,6 +67,7 @@ static const char *rcsid = "tftp-hpa $Id: tftpd.c,v 1.7 2001/03/30 01:30:50 hpa
 #include <stdio.h>
 #include <errno.h>
 #include <ctype.h>
+#define __USE_GNU		/* Necessary for basename() on glibc systems */
 #include <string.h>
 #include <stdlib.h>
 #include <pwd.h>
@@ -76,6 +78,15 @@ static const char *rcsid = "tftp-hpa $Id: tftpd.c,v 1.7 2001/03/30 01:30:50 hpa
 #include "tftpsubs.h"
 #include "recvfrom.h"
 
+#ifdef HAVE_TCPWRAPPERS
+#include <tcpd.h>
+
+int deny_severity	= LOG_WARNING;
+int allow_severity	= -1;	/* Don't log at all */
+
+struct request_info wrap_request;
+#endif
+
 void bsd_signal(int, void (*)(int));
 
 #ifndef HAVE_SIGSETJMP
@@ -144,7 +155,7 @@ struct options {
 static void
 usage(void)
 {
-	syslog(LOG_ERR, "Usage: %s [-cs] [-r option...] [directory ...]",
+	syslog(LOG_ERR, "Usage: %s [-c] [-u user] [-t timeout] [-r option...] [-s] [directory ...]",
 	       __progname);
 	exit(1);
 }
@@ -160,12 +171,16 @@ main(int argc, char **argv)
 	int on = 1;
 	int fd = 0;
 	int pid;
-	int i, j;
 	int c;
+	int setrv;
+	int timeout = 900;	/* Default timeout */
+	char *user = "nobody";	/* Default user */
 
-	openlog("tftpd", LOG_PID | LOG_NDELAY, LOG_DAEMON);
+	__progname = basename(argv[0]);
 
-	while ((c = getopt(argc, argv, "csr:")) != -1)
+	openlog(__progname, LOG_PID | LOG_NDELAY, LOG_DAEMON);
+
+	while ((c = getopt(argc, argv, "csu:r:t:")) != -1)
 		switch (c) {
 		case 'c':
 			cancreate = 1;
@@ -173,6 +188,12 @@ main(int argc, char **argv)
 		case 's':
 			secure = 1;
 			break;
+		case 't':
+		  timeout = atoi(optarg);
+		  break;
+		case 'u':
+		  user = optarg;
+		  break;
 		case 'r':
 		  for ( opt = options ; opt->o_opt ; opt++ ) {
 		    if ( !strcasecmp(optarg, opt->o_opt) ) {
@@ -220,82 +241,113 @@ main(int argc, char **argv)
 		}
 	}
 
-	pw = getpwnam("nobody");
+	pw = getpwnam(user);
 	if (!pw) {
-		syslog(LOG_ERR, "no nobody: %m");
+		syslog(LOG_ERR, "no user %s: %m", user);
 		exit(1);
 	}
 
+	if (ioctl(fd, FIONBIO, &on) < 0) {
+		syslog(LOG_ERR, "ioctl(FIONBIO): %m");
+		exit(1);
+	}
+
+	/* This means we don't want to wait() for children */
+	bsd_signal(SIGCHLD, SIG_IGN);
+
+	do {
+	  fd_set readset;
+	  struct timeval tv_timeout;
+	  
+
+	  FD_ZERO(&readset);
+	  FD_SET(fd, &readset);
+	  tv_timeout.tv_sec = timeout;
+	  tv_timeout.tv_usec = 0;
+
+	  if ( select(fd+1, &readset, NULL, NULL, &tv_timeout) == 0 )
+	    exit(0);		/* Timeout, return to inetd */
+
+	  fromlen = sizeof (from);
+	  n = myrecvfrom(fd, buf, sizeof (buf), 0,
+			 (struct sockaddr *)&from, &fromlen,
+			 &myaddr);
+
+	  if (n < 0) {
+	    syslog(LOG_ERR, "recvfrom: %m");
+	    exit(1);
+	  }
+
+#ifdef HAVE_TCPWRAPPERS
+	/* Verify if this was a legal request for us.  This has to be
+	   done before the chroot, while /etc is still accessible. */
+	  request_init(&wrap_request,
+		       RQ_DAEMON, __progname,
+		       RQ_FILE, fd,
+		       RQ_CLIENT_SIN, &from,
+		       RQ_SERVER_SIN, &myaddr,
+		       0);
+	  sock_methods(&wrap_request);
+	  if ( hosts_access(&wrap_request) == 0 ) {
+	    if ( deny_severity != -1 )
+	      syslog(deny_severity, "connection refused from %s",
+		     inet_ntoa(from.sin_addr));
+	    exit(1);		/* Access denied */
+	  } else if ( allow_severity != -1 ) {
+	    syslog(allow_severity, "connect from %s",
+		   inet_ntoa(from.sin_addr));
+	  }
+#endif
+
+	/*
+	 * Now that we have read the message out of the UDP
+	 * socket, we fork and go back to listening to the
+	 * socket.
+	 */
+	  pid = fork();
+	  if (pid < 0) {
+	    syslog(LOG_ERR, "fork: %m");
+	    exit(1);		/* Return to inetd, just in case */
+	  }
+	} while ( pid > 0 );	/* Parent process continues... */
+
+	/* Child process: handle the actual request here */
+
+	/* Chroot and drop privileges */
+
 	if (secure && chroot(".")) {
 		syslog(LOG_ERR, "chroot: %m");
 		exit(1);
 	}
 
-	(void) setegid(pw->pw_gid);
-	(void) setgid(pw->pw_gid);
-	(void) seteuid(pw->pw_uid);
-	(void) setuid(pw->pw_uid);
+#ifdef HAVE_SETREGID
+	setrv = setregid(pw->pw_gid, pw->pw_gid);
+#else
+	setrv = setegid(pw->pw_gid) || setgid(pw->pw_gid);
+#endif
 
-	if (ioctl(fd, FIONBIO, &on) < 0) {
-		syslog(LOG_ERR, "ioctl(FIONBIO): %m");
-		exit(1);
+#ifdef HAVE_SETREUID
+	setrv = setrv || setreuid(pw->pw_uid, pw->pw_uid);
+#else
+	/* Important: setuid() must come first */
+	setrv = setrv || setuid(pw->pw_uid) ||
+	  (geteuid() != pw->pw_uid && seteuid(pw->pw_uid));
+#endif
+
+	if ( setrv ) {
+	  syslog(LOG_ERR, "cannot drop privileges: %m");
+	  exit(1);
 	}
-	fromlen = sizeof (from);
-	n = myrecvfrom(fd, buf, sizeof (buf), 0,
-		       (struct sockaddr *)&from, &fromlen,
-		       &myaddr);
-	if (n < 0) {
-		syslog(LOG_ERR, "recvfrom: %m");
-		exit(1);
-	}
-	/*
-	 * Now that we have read the message out of the UDP
-	 * socket, we fork and exit.  Thus, inetd will go back
-	 * to listening to the tftp port, and the next request
-	 * to come in will start up a new instance of tftpd.
-	 *
-	 * We do this so that inetd can run tftpd in "wait" mode.
-	 * The problem with tftpd running in "nowait" mode is that
-	 * inetd may get one or more successful "selects" on the
-	 * tftp port before we do our receive, so more than one
-	 * instance of tftpd may be started up.  Worse, if tftpd
-	 * break before doing the above "recvfrom", inetd would
-	 * spawn endless instances, clogging the system.
-	 */
-	for (i = 1; i < 20; i++) {
-		pid = fork();
-		if (pid < 0) {
-			sleep(i);
-			/*
-			 * flush out to most recently sent request.
-			 *
-			 * This may drop some request, but those
-			 * will be resent by the clients when
-			 * they timeout.  The positive effect of
-			 * this flush is to (try to) prevent more
-			 * than one tftpd being started up to service
-			 * a single request from a single client.
-			 */
-			j = sizeof from;
-			i = myrecvfrom(fd, buf, sizeof (buf), 0,
-			    (struct sockaddr *)&from, &j, &myaddr);
-			if (i > 0) {
-				n = i;
-				fromlen = j;
-			}
-		} else
-			break;
-	}
-	if (pid < 0) {
-		syslog(LOG_ERR, "fork: %m");
-		exit(1);
-	} else if (pid != 0)
-		exit(0);
+
+	/* Close file descriptors we don't need */
 
 	from.sin_family = AF_INET;
 	alarm(0);
 	close(fd);
 	close(1);
+
+	/* Process the request... */
+
 	peer = socket(AF_INET, SOCK_DGRAM, 0);
 	if (peer < 0) {
 		syslog(LOG_ERR, "socket: %m");
